@@ -4,9 +4,15 @@ package Inline::Java::JNI ;
 
 use strict ;
 
-$Inline::Java::JNI::VERSION = '0.20' ;
+$Inline::Java::JNI::VERSION = '0.31' ;
 
+use DynaLoader ;
 use Carp ;
+use File::Basename ;
+
+
+# A place to attach the Inline object that is currently in Java land
+$Inline::Java::JNI::INLINE_HOOK = undef ;
 
 
 eval {
@@ -17,5 +23,4 @@ if ($@){
 }
 
 
-
 1 ;
