@@ -3,7 +3,7 @@ package Inline::Java::Protocol ;
 
 use strict ;
 
-$Inline::Java::Protocol::VERSION = '0.30' ;
+$Inline::Java::Protocol::VERSION = '0.31' ;
 
 use Inline::Java::Object ;
 use Inline::Java::Array ;
@@ -24,6 +24,17 @@ sub new {
 }
 
 
+sub ServerType {
+	my $this = shift ;
+
+	Inline::Java::debug("getting server type") ;
+
+	my $data = "server_type" ;
+
+	return $this->Send($data, 1) ;
+}
+
+
 sub Report {
 	my $this = shift ;
 	my $classes = shift ;
@@ -213,6 +224,7 @@ sub ValidateMember {
 sub ValidateArgs {
 	my $this = shift ;
 	my $args = shift ;
+	my $callback = shift ;
 
 	my @ret = () ;
 	foreach my $arg (@{$args}){
@@ -221,7 +233,12 @@ sub ValidateArgs {
 		}
 		elsif (ref($arg)){
 			if ((! UNIVERSAL::isa($arg, "Inline::Java::Object"))&&(! UNIVERSAL::isa($arg, "Inline::Java::Array"))){
-				croak "A Java method or member can only have Java objects, Java arrays or scalars as arguments" ;
+				if (! $callback){
+					croak "A Java method or member can only have Java objects, Java arrays or scalars as arguments" ;
+				}
+				else{
+					croak "A Java callback function can only return Java objects, Java arrays or scalars" ;
+				}
 			}
 
 			my $obj = $arg ;
@@ -259,23 +276,44 @@ sub Send {
 	my $data = shift ;
 	my $const = shift ;
 
-	my $resp = Inline::Java::get_JVM()->process_command($data) ;
+	my $inline = Inline::Java::get_INLINE($this->{module}) ;
+	my $resp = Inline::Java::__get_JVM()->process_command($inline, $data) ;
 
 	if ($resp =~ /^error scalar:([\d.]*)$/){
 		my $msg = pack("C*", split(/\./, $1)) ;
 		Inline::Java::debug("  packet recv error: $msg") ;
 		croak $msg ;
 	}
-	elsif ($resp =~ /^ok scalar:([\d.]*)$/){
+	elsif ($resp =~ s/^ok //){
+		return $this->DeserializeObject($const, $resp) ;
+	}
+
+	croak "Malformed response from server: $resp" ;
+}
+
+
+sub DeserializeObject {
+	my $this = shift ;
+	my $const = shift ;
+	my $resp = shift ;
+
+	if ($resp =~ /^scalar:([\d.]*)$/){
 		return pack("C*", split(/\./, $1)) ; 
 	}
-	elsif ($resp =~ /^ok undef:$/){
+	elsif ($resp =~ /^undef:$/){
 		return undef ;
 	}
-	elsif ($resp =~ /^ok object:(\d+):(.*)$/){
+	elsif ($resp =~ /^object:([01]):(\d+):(.*)$/){
 		# Create the Perl object wrapper and return it.
-		my $id = $1 ;
-		my $class = $2 ;
+		my $thrown = $1 ;
+		my $id = $2 ;
+		my $class = $3 ;
+
+		if ($thrown){
+			# If we receive a thrown object, we jump out of 'constructor
+			# mode' and process the returned object.
+			$const = 0 ;
+		}
 
 		if ($const){
 			$this->{obj_priv}->{java_class} = $class ;
@@ -296,16 +334,23 @@ sub Send {
 				$elem_class = $d[2] ;
 			}
 
-			my $perl_class = Inline::Java::java2perl($pkg, $elem_class) ;
-			if (Inline::Java::Class::ClassIsReference($elem_class)){
-				if (! Inline::Java::known_to_perl($pkg, $elem_class)){
-					if ($inline->get_java_config('AUTOSTUDY')){
-						$inline->_study([$elem_class]) ;
-					}
-					else{
-						$perl_class = "Inline::Java::Object" ;
-					}
-			 	}
+
+			my $perl_class = "Inline::Java::Object" ;
+			if ($elem_class){
+				# We have a real class or an array of real classes
+				$perl_class = Inline::Java::java2perl($pkg, $elem_class) ;
+				if (Inline::Java::Class::ClassIsReference($elem_class)){
+					if (! Inline::Java::known_to_perl($pkg, $elem_class)){
+						if (($thrown)||($inline->get_java_config('AUTOSTUDY'))){
+							$inline->_study([$elem_class]) ;
+						}
+						else{	
+							# Object is not known to Perl, it lives as a 
+							# Inline::Java::Object
+							$perl_class = "Inline::Java::Object" ;
+						}
+				 	}
+				}
 			}
 			else{
 				# We should only get here if an array of primitives types
@@ -323,10 +368,25 @@ sub Send {
 				$obj = $perl_class->__new($class, $inline, $id) ;
 			}
 
-			Inline::Java::debug("returning stub...") ;
-			return $obj ;
+			if ($thrown){
+				Inline::Java::debug("throwing stub...") ;
+				my ($msg, $score) = $obj->__isa('InlineJavaPerlCaller$PerlException') ;
+				if ($msg){
+					die $obj ;
+				}
+				else{
+					die $obj->GetObject() ;
+				}
+			}
+			else{
+				Inline::Java::debug("returning stub...") ;
+				return $obj ;
+			}
 		}
 	}
+	else{
+		croak "Malformed response from server: $resp" ;
+	}
 }
 
 
@@ -350,11 +410,11 @@ __DATA__
 	the request type and then we proceed to serve it.
 */
 class InlineJavaProtocol {
-	InlineJavaServer ijs ;
-	InlineJavaClass ijc ;
-	InlineJavaArray ija ;
-	String cmd ;
-	String response ;
+	private InlineJavaServer ijs ;
+	private InlineJavaClass ijc ;
+	private InlineJavaArray ija ;
+	private String cmd ;
+	private String response ;
 
 	InlineJavaProtocol(InlineJavaServer _ijs, String _cmd) {
 		ijs = _ijs ;
@@ -381,6 +441,9 @@ class InlineJavaProtocol {
 		else if (c.equals("get_member")){
 			GetJavaMember(st) ;
 		}		
+		else if (c.equals("server_type")){
+			ServerType(st) ;
+		}
 		else if (c.equals("report")){
 			Report(st) ;
 		}
@@ -472,6 +535,11 @@ class InlineJavaProtocol {
 	}
 
 
+	void ServerType(StringTokenizer st) throws InlineJavaException {
+		SetResponse(ijs.GetType()) ;
+	}
+
+
 	void ISA(StringTokenizer st) throws InlineJavaException {
 		String class_name = st.nextToken() ;
 		Class c = ijc.ValidateClass(class_name) ;
@@ -500,8 +568,19 @@ class InlineJavaProtocol {
 			Object p[] = (Object [])f.get(1) ;
 			Class clist[] = (Class [])f.get(2) ;
 
-			Object o = CreateObject(c, p, clist) ;
-			SetResponse(o) ;
+			try {
+				Object o = CreateObject(c, p, clist) ;
+				SetResponse(o) ;
+			}
+			catch (InlineJavaInvocationTargetException ite){
+				Throwable t = ite.GetThrowable() ;
+				if (t instanceof InlineJavaPerlCaller.InlineJavaException){
+					throw ((InlineJavaPerlCaller.InlineJavaException)t).GetException() ;
+				}
+				else{
+					SetResponse(new InlineJavaServerThrown(t)) ;
+				}
+			}
 		}
 		else{
 			// Here we send the type of array we want, but CreateArray
@@ -568,8 +647,13 @@ class InlineJavaProtocol {
 				Throwable t = e.getTargetException() ;
 				String type = t.getClass().getName() ;
 				String msg = t.getMessage() ;
-				throw new InlineJavaException(
-					"Method " + name + " in class " + class_name + " threw exception " + type + ": " + msg) ;
+				ijs.debug("Method " + name + " in class " + class_name + " threw exception " + type + ": " + msg) ;
+				if (t instanceof InlineJavaPerlCaller.InlineJavaException){
+					throw ((InlineJavaPerlCaller.InlineJavaException)t).GetException() ;
+				}
+				else{
+					SetResponse(new InlineJavaServerThrown(t)) ;
+				}
 			}
 		}
 	}
@@ -718,8 +802,9 @@ class InlineJavaProtocol {
 			Throwable t = e.getTargetException() ;
 			String type = t.getClass().getName() ;
 			String msg = t.getMessage() ;
-			throw new InlineJavaException(
-				"Constructor for class " + name + " with signature " + ijs.CreateSignature(proto) + " threw exception " + type + ": " + msg) ;
+			throw new InlineJavaInvocationTargetException(
+				"Constructor for class " + name + " with signature " + ijs.CreateSignature(proto) + " threw exception " + type + ": " + msg,
+				t) ;
 		}
 
 		return ret ;
@@ -840,7 +925,6 @@ class InlineJavaProtocol {
 				if (type.equals(t)){
 					ijs.debug("  has matching type " + t) ;
 					fl.add(fl.size(), f) ;
-					break ;
 				}
 			}
 		}
@@ -852,10 +936,13 @@ class InlineJavaProtocol {
 				"Member " + name + " of type " + type + " for class " + c.getName() +
 					" not found") ;
 		}
-		else if (fl.size() == 1){
+		else {
 			// Now we need to force the arguments received to match
 			// the methods signature.
-			Field f = (Field)fl.get(0) ;
+
+			// If we have more that one, we use tha last one, which is the most
+			// specialized
+			Field f = (Field)fl.get(fl.size() - 1) ;
 			param = f.getType() ;
 
 			String msg = "For member " + name + " of class " + c.getName() + ": " ;
@@ -881,27 +968,38 @@ class InlineJavaProtocol {
 		script
 	*/
 	void SetResponse (Object o) throws InlineJavaException {
+		response = "ok " + SerializeObject(o) ;
+	}
+
+
+	String SerializeObject(Object o) throws InlineJavaException {
 		if (o == null){
-			response = "ok undef:" ;
+			return "undef:" ;
 		}
 		else if ((ijc.ClassIsNumeric(o.getClass()))||(ijc.ClassIsChar(o.getClass()))||(ijc.ClassIsString(o.getClass()))){
-			response = "ok scalar:" + unpack(o.toString()) ;
+			return "scalar:" + unpack(o.toString()) ;
 		}
 		else if (ijc.ClassIsBool(o.getClass())){
 			String b = o.toString() ;
-			response = "ok scalar:" + unpack((b.equals("true") ? "1" : "0")) ;
+			return "scalar:" + unpack((b.equals("true") ? "1" : "0")) ;
 		}
 		else {
 			// Here we need to register the object in order to send
 			// it back to the Perl script.
+			boolean thrown = false ;
+			if (o instanceof InlineJavaServerThrown){ 
+				thrown = true ;
+				o = ((InlineJavaServerThrown)o).GetThrowable() ;
+			}			
 			int id = ijs.objid ;
 			ijs.PutObject(id, o) ;
-			response = "ok object:" + String.valueOf(id) +
+			return "object:" + (thrown ? "1" : "0") + ":" + String.valueOf(id) +
 				":" + o.getClass().getName() ;
 		}
 	}
 
 
+
 	/* Equivalent to Perl pack */
 	public String pack(String s){
 		StringTokenizer st = new StringTokenizer(s, ".") ;
